﻿
using System;
using System.Collections.Generic;
using System.Linq;
using Sandbox.Common.ObjectBuilders;
using Sandbox.Definitions;
using Sandbox.ModAPI;
using Sandbox.ModAPI.Interfaces;
using VRage;
using VRageMath;

namespace Scripts.KSWH
{
    /// <summary>
    /// Abstract class with functions the AI must have and all types of AI's always need. Also holds a few key variables for the AI to function
    /// </summary>
    public abstract class Ai : Controlable
    {
        protected IMyControllableEntity m_cockpit;
        public int id { private set; get; }
        public IMyEntity Target;
        public int Range;
        protected bool UsedEcm;
        protected bool IsSpotted = false;
        protected bool IsAttacking = false;
        public bool IsControlable { protected set; get; }
        protected int Count = 0;
        protected int MaxCount;
        protected bool SelfDestructing = false;
        private DateTime mLastGunTrigger = DateTime.Now;
        private bool shooting;
        private FastResourceLock mFireLock = new FastResourceLock();
        private bool mSetSpawnDate;
        private DateTime mSpawnDate;
        private bool doUpdate = true;
        private FastResourceLock executionLock = new FastResourceLock();
        private bool executing = false;
        protected Navigator nav;
        private bool broadcasting = false;

        /// <summary>
        /// Grant execution time to the ship so it can setup a background worker if needed to plot its course
        /// </summary>
        public void GrantExecutionTime()
        {
            if (ExecutionCheck())
            {
                return; //busy
            }
            if (!IsControlable)
            {
                MissionComponent.RemoveShipFromAiHook(id);
                return;
            }
            ExecutionCheck(true, true);
            MyAPIGateway.Parallel.StartBackground(UpdateWrapper);

            ExecutionCheck(true); //remove busy state

        }


        /// <summary>
        /// wrapper that will handle the dispose if the ship is no longer usable
        /// </summary>
        private void UpdateWrapper()
        {
            MyApiWrapper.GetAiLock("ai running update");
            try
            {
                Update();
            }
            catch (Exception)
            {
                if (MissionComponent.writeDebug)
                {
                    MissionComponent.Speak("Error", "Ship had a thread issue during update");
                }
            }
            finally
            {
                MyApiWrapper.ReleaseAiLock();
            }
            
            
            bool control = IsControlable;
            if (control)
            {
                MissionComponent.ReQueueAiHook(this);
                return;
            }
            if (MissionComponent.writeDebug)
            {
                MissionComponent.Speak("Ai", "ship destoryed");
            }
            if (nav != null)
            {
                MissionComponent.AutoHandler.RemoveNavigator(Grid);
                nav.myGrid.Close();
                nav = null;
            }
            MissionComponent.RemoveShipFromAiHook(id);
        }

        /// <summary>
        /// PUt in the tmpId for this ship (easy refer back toe Dictionary)
        /// </summary>
        /// <param name="tmpId">int with the tmpId</param>
        protected void SetId(int tmpId)
        {
            id = tmpId;
        }

        /// <summary>
        /// Check if the AI entitiy already has calculation power
        /// </summary>
        /// <returns></returns>
        private bool ExecutionCheck(bool trySet = false, bool to = false)
        {
            executionLock.AcquireExclusive();
            bool ret;
            if (trySet)
            {
                executing = to;
                ret = to;
            }
            else
            {
                ret = executing;
            }  
            executionLock.ReleaseExclusive();
            return ret;
        }

        public void CheckRadar()
        {
            
        } 

        public virtual void Update()
        {
            if (!doUpdate)
            {
                doUpdate = true;
                return;
            }
            doUpdate = false;
            m_cockpit = GetCockpit();
            if (m_cockpit == null)
            {
                IsControlable = false;
                SelfDestruct();
                return;
            }
            IsControlable = true;

            var tmp = new List<IMySlimBlock>();
            //check for power

            Grid.GetBlocks(tmp,
                x => x.FatBlock != null && x.FatBlock.BlockDefinition.TypeIdString == "MyObjectBuilder_Reactor");

            if (tmp.Count == 0)
            {
                if (MissionComponent.writeDebug)
                {
                    MissionComponent.Speak("AI HOOK", Grid.DisplayName + " has no Reactors");
                }
                IsControlable = false;
                return; //no power, ship no longer controlable
            }

            //check for thrusters
            tmp.Clear();
            Grid.GetBlocks(tmp, x => x.FatBlock != null && x.FatBlock.BlockDefinition.TypeIdString == "MyObjectBuilder_Gyro");

            if (tmp.Count == 0)
            {
                if (MissionComponent.writeDebug)
                {
                    MissionComponent.Speak("AI HOOK", Grid.DisplayName + " has no gyro's");
                }
                IsControlable = false;
                return; //no gyroscopes
            }

            //check for thrusters
            tmp.Clear();
            Grid.GetBlocks(tmp, x => x.FatBlock != null && x.FatBlock.BlockDefinition.TypeIdString == "MyObjectBuilder_Thrust");

            if (tmp.Count == 0)
            {
                if (MissionComponent.writeDebug)
                {
                    MissionComponent.Speak("AI HOOK", Grid.DisplayName + " has no thrusters");
                }
                IsControlable = false;
                return; //no thrusters
            }

            if (!mSetSpawnDate)
            {
                mSpawnDate = DateTime.Now;
                mSetSpawnDate = true;
            }
            else
            {
                TimeSpan span = (DateTime.Now - mSpawnDate);
                if (span.TotalMinutes > GetLivespawn())
                {
                    if (MissionComponent.writeDebug)
                    {
                        MissionComponent.Speak("AI HOOK", Grid.DisplayName + " out of livetime");
                    }
                    SelfDestruct();
                    IsControlable = false;
                }
            }
            CheckBeaconStatus();
        }

        protected abstract void UseEcm();

        /// <summary>
        /// Set the ownsership of a entity (faction level)
        /// </summary>
        /// <param name="entity">the entity (like a ship)</param>
        /// <param name="tag">tag of the faction</param>
        protected void SetOwsnershipToEnemy(IMyCubeGrid entity)
        {
            entity.ChangeGridOwnership(MissionComponent.host.enemy, MyOwnershipShareModeEnum.Faction);

        }

        public void Dispose()
        {
            Grid = null;
            Target = null;
            m_cockpit = null;
            IsControlable = false;
        }

        /// <summary>
        /// Function that removes all overrides on gyro's if they are on there from a template. 
        /// </summary>
        protected void DisableOverride()
        {
            //MissionComponent.Speak("AI_controller", "disabling all overwrites");
            var lst = new List<IMySlimBlock>();
            MyDefinitionId gyroDef = new MyDefinitionId(typeof(MyObjectBuilder_Gyro));

            Grid.GetBlocks(lst, (x) => x.FatBlock != null && x.FatBlock.BlockDefinition.SubtypeId.Contains("Gyro") && x.FatBlock.BlockDefinition.TypeId.Equals(gyroDef.TypeId));
            if (lst.Count < 1)
            {
                SelfDestruct();
                return;
            }
            var gyro = lst[0];
            if (gyro == null)
            {
                MissionComponent.Speak("System", "List entity is not filled");
                return;
            }
            var actions = new List<ITerminalAction>();
            MyAPIGateway.TerminalActionsHelper.GetActions(gyro.FatBlock.GetType(), actions, (x) => x.Id.Contains("Override"));
            var action = actions[0];
            if (action == null)
            {
                return;
            }
            //MissionComponent.Speak("system", "Overwriting " + lst.Count + " blocks");

            foreach (var block in lst)
            {
                action.Apply(block.FatBlock);
            }
        }

        /// <summary>
        /// Arm all warheads and destroy yourself
        /// </summary>
        protected void SelfDestruct()
        {
            if (SelfDestructing)
            {
                return;
            }
            if (MissionComponent.writeDebug)
            {
                MissionComponent.Speak("AI", "Activating self destruct");
            }
            SelfDestructing = true;
            //MissionComponent.Speak("AI", "Selfdestructing");
            var lst = new List<IMySlimBlock>();
            MyDefinitionId nukeDef = new MyDefinitionId(typeof(MyObjectBuilder_Warhead));
            Grid.GetBlocks(lst, x => x.FatBlock != null && x.FatBlock.BlockDefinition.SubtypeId.Contains("Warhead") && x.FatBlock.BlockDefinition.TypeId.Equals(nukeDef.TypeId));
            if (lst.Count <= 0)
            {
                //MissionComponent.Speak("AI", "no nukes found");
                return; //no nukes, no way you can selfdestruct
            }
            var nuke = lst[0];
            var actions = new List<ITerminalAction>();
            MyAPIGateway.TerminalActionsHelper.GetActions(nuke.FatBlock.GetType(), actions, x => x.Id.Contains("Start"));
            if (actions.Count == 0)
            {
                //MissionComponent.Speak("AI", "no nuke action found");
                return; //no action found
            }
            var action = actions.First();
            foreach (var block in lst)
            {
                //MissionComponent.Speak("AI", "starting nuke countdown");
                action.Apply(block.FatBlock);
            }
            IsControlable = false;

            if (this is Missile)
            {
                MissionComponent.AlarmHandler.MissleGone();
            }
            else
            {
                MissionComponent.AlarmHandler.Clear();
            }
        }

        /// <summary>
        /// Check if the ai should turn it beacon on, this method respects the radar ranges
        /// </summary>
        private void CheckBeaconStatus()
        {
            if (MissionComponent.rHandler.InRange(Grid.GetPosition()))
            {
                if (broadcasting)
                {
                    return;
                }
                //in area with radar couverage, turn on the beacon
                SetBeacon("OnOff_On");
                broadcasting = true;
            }
            else
            {
                if (broadcasting)
                {
                    //turn beacon off
                    SetBeacon("OnOff_Off");
                    broadcasting = false;
                }
            }
        }

        private void SetBeacon(string statusTag)
        {
            var tmp = new List<IMySlimBlock>();
            //check for power
            Grid.GetBlocks(tmp,
                x => x.FatBlock != null && x.FatBlock.BlockDefinition.TypeIdString == "MyObjectBuilder_Beacon");
            if (tmp.Count != 0)
            {
                var actions = new List<ITerminalAction>();
                MyAPIGateway.TerminalActionsHelper.GetActions(tmp.First().FatBlock.GetType(), actions,
                    x => x.Id.Equals(statusTag));
                if (actions.Count == 0)
                {
                    if (MissionComponent.writeDebug)
                    {
                        MissionComponent.Speak("ShipAi", "No action found");
                    }
                    return;
                }
                foreach (IMySlimBlock block in tmp)
                {
                    actions.First().Apply(block.FatBlock);
                }
            }
            else
            {
                if (MissionComponent.writeDebug)
                {
                    MissionComponent.Speak("shipAi", "No beacon found");  
                }
            }
        }

        protected bool FireMountedGunsStart(bool start)
        {
            bool succes = mFireLock.TryAcquireExclusive();
            if (!succes)
            {
                return start;
            }
            TimeSpan span = (DateTime.Now - mLastGunTrigger);
            if (span.TotalSeconds <= 1)
            {
                if (MissionComponent.writeDebug)
                {
                    MissionComponent.Speak("Ai guns", "Not firing, timespawn to short " + span.TotalSeconds);
                }
                mFireLock.ReleaseExclusive();
                return !start; //thread savety thing
            }

            if (start && shooting)
            {
                mFireLock.ReleaseExclusive();
                return start; // already shooting
            }
            if (start)
            {
                shooting = true;
            }
            try
            {
                MyDefinitionId gunDef = new MyDefinitionId(typeof (MyObjectBuilder_SmallGatlingGun));
                var lst = new List<IMySlimBlock>();
                Grid.GetBlocks(lst, x => x.FatBlock != null && x.FatBlock.BlockDefinition.TypeId.Equals(gunDef.TypeId));
                if (lst.Count < 0)
                {
                    mFireLock.ReleaseExclusive();
                    return !start; //no guns to fire
                }
                var gun = lst[0];
                var actions = new List<ITerminalAction>();
                string actionTag = start ? "Shoot_On" : "Shoot_Off";
                mLastGunTrigger = DateTime.Now;
                MyAPIGateway.TerminalActionsHelper.GetActions(gun.FatBlock.GetType(), actions,
                    x => x.Id.Equals(actionTag));

                if (actions.Count == 0)
                {
                    if (MissionComponent.writeDebug)
                    {
                        MissionComponent.Speak("ai", "no gun action found");
                    }
                    mFireLock.ReleaseExclusive();
                    return !start;
                }
                var action = actions.First();
                ThreadWrapper.ListAction(lst,action);
            }
            catch (Exception)
            {
                //meh
                MissionComponent.Speak("ai", "Failed to fire gun");
            }
            mFireLock.ReleaseExclusive();
            shooting = start;
            return start;
        }

        protected abstract int GetLivespawn();

        protected static IMyEntity GetPlayerInSphere(ref BoundingSphere sphere)
        {
            HashSet<IMyEntity> ents = new HashSet<IMyEntity>();
            MyAPIGateway.Entities.GetEntities(ents, x => x != null);
            return (from entity in ents let controllable = entity as IMyControllableEntity where controllable != null && (MyAPIGateway.Multiplayer.Players.GetPlayerControllingEntity(entity) != null || controllable == MyAPIGateway.Session.ControlledObject) select entity).FirstOrDefault();
        }

        protected Vector3D? GetTargetDestination(Vector3D targetLoc, int minOffset, int maxOffset)
        {
            Random r = new Random(DateTime.Now.Second);
            double xAuto = targetLoc.X + r.Next(-400, 400);
            double yAuto = targetLoc.Y + r.Next(-400, 400);
            double zAuto = targetLoc.Z + r.Next(-400, 400);
            Vector3D? ret = new Vector3D(xAuto, yAuto, zAuto);
            return ret;
        }
    }
}
