﻿using System;
using Sandbox.Common.ObjectBuilders;
using Sandbox.ModAPI;
using Sandbox.ModAPI.Interfaces;
using System.Collections.Generic;
using System.Linq;
using System.Security.Policy;
using Sandbox.Definitions;
using VRage;
using VRageMath;
using IMySlimBlock = Sandbox.ModAPI.Ingame.IMySlimBlock;

namespace Scripts.KSWH
{
    /// <summary>
    /// Class that will handle the alarm entities in the world
    /// </summary>
    internal class Alarm
    {
        private int Updater = 0;
        private int lastAlarmCount = 0;
        private int IncommingMissles = 0;
        private readonly List<Compass> Compasses;
        private Station PointTarget;

        private List<IMySlimBlock> AlarmBlocks;
        private List<IMySlimBlock> KillBlocks;
        private List<IMySlimBlock> DamageBlocks;
        private List<IMySlimBlock> RadationBlocks;
        private List<IMySlimBlock> SignalBlock;
        private List<IMySlimBlock> ClearBlocks;
        private List<IMySlimBlock> BalisticBlocks;
        private List<IMySlimBlock> BclearBlocks;
        private List<IMySlimBlock> EcmBlocks;

        private FastResourceLock AlarmLock;

        private int alarmCount;

        private ITerminalAction startAction;

        private enum AlarmType
        {
            Alarm,
            Kill,
            Damage,
            Radation,
            Signal,
            Clear,
            Balistic,
            BalisticClear,
            Ecm
        }

        /// <summary>
        /// default constructor
        /// </summary>
        public Alarm()
        {
            //init all alarm block lists and lock
            AlarmBlocks = new List<IMySlimBlock>();
            KillBlocks = new List<IMySlimBlock>();
            DamageBlocks = new List<IMySlimBlock>();
            RadationBlocks = new List<IMySlimBlock>();
            SignalBlock = new List<IMySlimBlock>();
            ClearBlocks = new List<IMySlimBlock>();
            BalisticBlocks = new List<IMySlimBlock>();
            BclearBlocks = new List<IMySlimBlock>();
            EcmBlocks = new List<IMySlimBlock>();
            AlarmLock = new FastResourceLock();
            //init compass list
            Compasses = new List<Compass>();
            //prepare alarm handler
            alarmCount = 0;
            ScanForAlarms();
            ScanForCompass();
        }

        public void Update()
        {
            Updater++;
            if (Updater >= 2500)
            {
                if (MissionComponent.writeDebug)
                {
                    MissionComponent.Speak("alarmhandler","Alarm handler scanning for new alarms");
                }
                Updater = 0;
                MyAPIGateway.Parallel.StartBackground(ScanForAlarms);
                ScanForCompass();
                FindPointTarget();
            }
            UpdateCompasses();
        }

        public void ScanNow()
        {
            Updater = 2500;
        }

        /// <summary>
        /// Scan the world for items that are market for a alarm system
        /// </summary>
        public void ScanForAlarms()
        {
            //todo: update for alarms 2.0
            AlarmLock.AcquireExclusive();
            ResetAllBlockLists();
            HashSet<IMyEntity> ents = new HashSet<IMyEntity>();
            MyAPIGateway.Entities.GetEntities(ents, x => x is IMyCubeGrid && x.DisplayName.Contains(StaticInformation.AlarmTag));
            int currentAlarmCount = 0;
            foreach (IMyEntity ent in ents)
            {
                List<Sandbox.ModAPI.IMySlimBlock> blocks = new List<Sandbox.ModAPI.IMySlimBlock>();
                var grid = ent as IMyCubeGrid;
                if (grid == null)
                {
                    continue;
                }
                grid.GetBlocks(blocks, x => x.FatBlock != null);
                foreach (Sandbox.ModAPI.IMySlimBlock block in blocks)
                {
                    string target = block.FatBlock.BlockDefinition.SubtypeId.Replace("_Large", "")
                        .Replace("_Small", "");
                    switch (target)
                    {
                        case StaticInformation.DamTag:
                            DamageBlocks.Add(block);
                            currentAlarmCount++;
                            break;
                        case StaticInformation.KillTag:
                            KillBlocks.Add(block);
                            currentAlarmCount++;
                            break;
                        case StaticInformation.NoBallisticsTag:
                            BclearBlocks.Add(block);
                            currentAlarmCount++;
                            break;
                        case StaticInformation.BallisticTag:
                            BalisticBlocks.Add(block);
                            currentAlarmCount++;
                            break;
                        case StaticInformation.RadTag:
                            RadationBlocks.Add(block);
                            currentAlarmCount++;
                            break;
                        case StaticInformation.EcmTag:
                            EcmBlocks.Add(block);
                            currentAlarmCount++;
                            break;
                        case StaticInformation.Sigtag:
                            SignalBlock.Add(block);
                            if (MissionComponent.AlarmDebug)
                            {
                                MissionComponent.Speak("alarm", "Adding signature alarm");
                            }
                            currentAlarmCount++;
                            break;
                        case StaticInformation.ClearTag:
                            ClearBlocks.Add(block);
                            currentAlarmCount++;
                            break;
                        case StaticInformation.AlarmTag:
                            AlarmBlocks.Add(block);
                            currentAlarmCount++;
                            break;
                        default:
                            if (MissionComponent.AlarmDebug)
                            {
                                MissionComponent.Speak("Alarm", "Cannot enlist " + target);
                            }
                            break;
                    }
                }
            }

            AlarmLock.ReleaseExclusive();
            if (MissionComponent.AlarmDebug)
            {
                MissionComponent.Speak("Alarm", "alarm block total: " + currentAlarmCount);
            }

            if (lastAlarmCount == currentAlarmCount) return;
            MissionComponent.Speak("base", "Now has " + currentAlarmCount + " alarms");
            lastAlarmCount = currentAlarmCount;
        }

        public void ScanForCompass()
        {
            Compasses.Clear();
            HashSet<IMyEntity> tmpCache = new HashSet<IMyEntity>();
            MyAPIGateway.Entities.GetEntities(tmpCache, (x) => x is IMyCubeGrid && x.DisplayName.Equals(StaticInformation.CompassTag));
            if (MissionComponent.writeDebug)
            {
                MissionComponent.Speak("alarmhandler", "found " + tmpCache.Count + " new Compasses");
            }
            foreach (IMyEntity com in tmpCache)
            {
                AddCompass(com);
            }
        }

        private void AddCompass(IMyEntity com)
        {
            Compasses.Add(new Compass(com as IMyCubeGrid));
            List<IMyEntity> parts = new List<IMyEntity>();
            com.GetChildren(parts);
            foreach (IMyEntity part in parts)
            {
                AddCompass(part);
            }
        }

        public void UpdateCompasses()
        {
            if (Compasses.Count == 0)
            {
                return; //no compasses, no need to do this
            }
            if (PointTarget == null)
            {
                return; //no target, no update needed
            }
            foreach (Compass com in Compasses)
            {
                com.Update(PointTarget.Grid.GetPosition());
            }
        }

        /// <summary>
        /// Activate a signature alarm
        /// </summary>
        public void NewSig()
        {
           StartAlarmBlock(AlarmType.Signal);
        }

        /// <summary>
        /// Start master alarm
        /// </summary>
        public void Alert(int priority, string message)
        {
            alarmCount++;
            StartAlarmBlock(AlarmType.Alarm);
            MissionComponent.hud.AddMessage(priority, message);
        }

        /// <summary>
        /// give all clear signal
        /// </summary>
        public void Clear()
        {
            KillSignal();
            alarmCount--;
            if (MissionComponent.writeDebug)
            {
                MissionComponent.Speak("AlarmSystem", "Lowered alarm level to " + alarmCount);
            }
            if (alarmCount <= 0)
            {
                StartAlarmBlock(AlarmType.Clear);
                alarmCount = 0;
            }
        }

        /// <summary>
        /// Give ecm warning
        /// </summary>
        public void EcmWarning()
        {
            StartAlarmBlock(AlarmType.Ecm);
        }

        /// <summary>
        /// Give radiation warning
        /// </summary>
        public void RadWarning()
        {
            StartAlarmBlock(AlarmType.Radation);
        }

        /// <summary>
        /// Start a firewall damage alarm
        /// </summary>
        public void DamageAlarm()
        {
            StartAlarmBlock(AlarmType.Damage);
        }

        /// <summary>
        /// Ring the ballistic alarm
        /// </summary>
        public void MissileWarning()
        {
            IncommingMissles++;
            StartAlarmBlock(AlarmType.Balistic);
            MissionComponent.hud.AddMessage((int)Hud.MessagePriority.Missile, Hud.MessageLead + "///WARNING::                                                                                                  \n" +
                                                 ">Incomming signature detected\n    -SIZE: small\n    -ARMOR: none\n    -OFFENSE: none\n>HIGH THREAT LEVEL\n\n***IMPACT IMMINENT***");
        }

        /// <summary>
        /// No more missle warning
        /// </summary>
        public void MissleGone()
        {
            IncommingMissles--;
            KillSignal();
            if (IncommingMissles <= 0)
            {
                IncommingMissles = 0;
                StartAlarmBlock(AlarmType.BalisticClear);
            }
            
        }

        public void KillSignal()
        {
            StartAlarmBlock(AlarmType.Kill);
        }

        private void FindPointTarget()
        {
            if (MissionComponent.writeDebug)
            {
                MissionComponent.Speak("compass", "Realigning");
            }
            if (PointTarget == null || !PointTarget.Usefull)
            { //no current target or current target is destoryed, select new one
                if (MissionComponent.StationHandler.Stations.Count == 0)
                {
                    //no hostile stations found, so nothing to point at
                    PointTarget = null;
                }
                //pick new target from stationhandler
                PointTarget =
                    MissionComponent.StationHandler.Stations.FirstOrDefault();
                if (MissionComponent.writeDebug)
                {
                    MissionComponent.Speak("Compass", "has target");
                }
            }

        }

        /// <summary>
        /// Simple method to clear all lists with alarm blocks
        /// </summary>
        private void ResetAllBlockLists()
        {
            AlarmBlocks.Clear();
            KillBlocks.Clear();
            DamageBlocks.Clear();
            RadationBlocks.Clear();
            SignalBlock.Clear();
            ClearBlocks.Clear();
            BalisticBlocks.Clear();
            BclearBlocks.Clear();
            EcmBlocks.Clear();
        }

        private void StartAlarmBlock(AlarmType type)
        {
            AlarmLock.AcquireShared();
            switch (type)
            {
                case AlarmType.Alarm:
                    foreach (var alarmBlock in AlarmBlocks)
                    {
                        StartTimerAction(alarmBlock);
                    }
                    break;
                case AlarmType.Kill:
                    foreach (var killBlock in KillBlocks)
                    {
                        StartTimerAction(killBlock);
                    }
                    break;
                case AlarmType.Damage:
                    foreach (var damageBlock in DamageBlocks)
                    {
                        StartTimerAction(damageBlock);
                    }
                    break;
                case AlarmType.Radation:
                    foreach (var radationBlock in RadationBlocks)
                    {
                        StartTimerAction(radationBlock);
                    }
                    break;
                case AlarmType.Signal:
                    foreach (var signalBlock in SignalBlock)
                    {
                        StartTimerAction(signalBlock);
                    }
                    break;
                case AlarmType.Clear:
                    foreach (var clearBlock in ClearBlocks)
                    {
                        StartTimerAction(clearBlock);
                    }
                    break;
                case AlarmType.Balistic:
                    foreach (var balisticBlock in BalisticBlocks)
                    {
                        StartTimerAction(balisticBlock);
                    }
                    break;
                case AlarmType.BalisticClear:
                    foreach (var bclearBlock in BclearBlocks)
                    {
                        StartTimerAction(bclearBlock);
                    }
                    break;
                case AlarmType.Ecm:
                    foreach (var ecmBlock in EcmBlocks)
                    {
                        StartTimerAction(ecmBlock);
                    }
                    break;
            }
            AlarmLock.ReleaseShared();
        }

        /// <summary>
        /// Apply the start action on a given timer block
        /// </summary>
        /// <param name="timer">The timer block that should start</param>
        private void StartTimerAction(IMySlimBlock timer)
        {
            try
            {
                if (startAction == null)
                {
                    var actions = new List<ITerminalAction>();
                    MyAPIGateway.TerminalActionsHelper.GetActions(timer.FatBlock.GetType(), actions, x => x.Id.Contains("Start"));
                    var action = actions.FirstOrDefault();
                    if (action == null)
                    {
                        MissionComponent.Speak("CRITICAL", "Action is NULL when loading reference");
                        return;
                    }
                    startAction = action;
                }
                startAction.Apply((IMyCubeBlock)timer.FatBlock);

            }
            catch (Exception)
            {
                //failsave to prevent carsh if the timer explodes while the action is being applied 
            }
        }

        /// <summary>
        /// Trigger a event (defined by tag) on a alarm or hq block
        /// </summary>
        /// <param name="tag">event that should be raised</param>
        /*private void StartTimer(string tag)
        {
            foreach (IMyCubeGrid tower in Towers)
            {
                List<Sandbox.ModAPI.IMySlimBlock> timers = new List<Sandbox.ModAPI.IMySlimBlock>();
                List<Sandbox.ModAPI.IMySlimBlock> timersSmall = new List<Sandbox.ModAPI.IMySlimBlock>();
                MyDefinitionId timer = new MyDefinitionId(typeof(MyObjectBuilder_TimerBlock), "TimerBlockLarge");
                MyDefinitionId tinyTimer = new MyDefinitionId(typeof(MyObjectBuilder_TimerBlock), "TimerBlockSmall");
                
                tower.GetBlocks(timers, x => x.FatBlock != null && x.FatBlock.BlockDefinition == timer && x.FatBlock.DisplayNameText.ToLower().Contains(tag));
                tower.GetBlocks(timersSmall, x => x.FatBlock != null && x.FatBlock.BlockDefinition == tinyTimer && x.FatBlock.DisplayNameText.ToLower().Contains(tag));
                timers.AddRange(timersSmall);
               // MissionComponent.Speak("base", "aplying action to " + timers.Count + " timers");
                foreach (IMySlimBlock t in timers)
                {
                    if (startAction == null)
                    {
                        var actions = new List<ITerminalAction>();
                        MyAPIGateway.TerminalActionsHelper.GetActions(t.FatBlock.GetType(), actions, x => x.Id.Contains("Start"));
                        var action = actions.FirstOrDefault();
                        if (action == null)
                        {
                            MissionComponent.Speak("CRITICAL", "Action is NULL when loading reference");
                            return;
                        }
                        startAction = action;
                    }
                   // MissionComponent.Speak("system", "triggering action on block " + t.FatBlock.DisplayNameText);
                    startAction.Apply((IMyCubeBlock)t.FatBlock);
                   
                }
            }
        }*/
    }
}
