﻿using System;
using System.CodeDom;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Sandbox.ModAPI;
using VRage;
using VRageMath;

namespace Scripts.KSWH
{
    class AutopilotHandler
    {
        private Dictionary<IMyCubeGrid, Navigator> allNavigators;
        private Queue<IMyCubeGrid> updateQueue;
        private static readonly int GridsPerPing = 1;
        private FastResourceLock QueueLock;
        private FastResourceLock NavigatorLock;
        private int updateCounter = 10;


        public AutopilotHandler()
        {
            QueueLock = new FastResourceLock();
            NavigatorLock = new FastResourceLock();
            updateQueue = new Queue<IMyCubeGrid>();  
            Settings.Open();
            allNavigators = new Dictionary<IMyCubeGrid, Navigator>();
        }

        /// <summary>
        /// Add a new navigator grid to the task list and queque in a threasave manner
        /// </summary>
        /// <param name="grid">The grid that should be autopiloted</param>
        /// <param name="destination">Location where the ship should be flown to</param>
        public Navigator AddNavigator(IMyCubeGrid grid, Vector3D? destination)
        {
            NavigatorLock.AcquireExclusive();
            if (!allNavigators.ContainsKey(grid))
            {
                StringBuilder sb = new StringBuilder();
                sb.Append('c');
                sb.Append(destination.Value.X);
                sb.Append(',');
                sb.Append(destination.Value.Y);
                sb.Append(',');
                sb.Append(destination.Value.Z);

                Navigator nav = new Navigator(grid, sb.ToString());
                nav.ForceOverwriteDestination(destination);
                allNavigators.Add(grid,nav);
                NavigatorLock.ReleaseExclusive();
                QueueLock.AcquireExclusive();
                updateQueue.Enqueue(grid);
                QueueLock.ReleaseExclusive();
                return nav;
            }
            
            //Key already exists, should not be readded
            MissionComponent.Speak("ERROR", "AutopilotHandler added the same navigator twice");
            var ret = allNavigators[grid];
            NavigatorLock.ReleaseExclusive();
            return ret;
            
        }

        /// <summary>
        /// Remove grid from navigators
        /// NOTE: this will NOT dequeque the item from updates. When the updater finds the removed item in the list it will skip it
        /// </summary>
        /// <param name="grid">Grid you want removed</param>
        public void RemoveNavigator(IMyCubeGrid grid)
        {
            NavigatorLock.AcquireExclusive();
            allNavigators.Remove(grid);
            NavigatorLock.ReleaseExclusive();
        }

        public void DoUpdate(bool mainThread = true)
        {
            if (updateCounter < 10)
            {
                updateCounter++;
                return;
            }
            updateCounter = 0;
            int cycles;
            QueueLock.AcquireExclusive();
            var queLength = updateQueue.Count;
            if (queLength == 0)
            {
                //nothing in que, no need to do anything
                QueueLock.ReleaseExclusive();
                return;
            }

            //calculate max cycles in this ping
            if (queLength < GridsPerPing)
            {
                cycles = queLength;
            }
            else
            {
                cycles = GridsPerPing;
            }
            List<Navigator> navs = new List<Navigator>();
            
            NavigatorLock.AcquireShared();
            for (int i = 0; i < cycles; i++)
            {
                var key = updateQueue.Dequeue();
                if (allNavigators.ContainsKey(key))
                {
                    navs.Add(allNavigators[key]);
                    updateQueue.Enqueue(key);
                }
            }
            NavigatorLock.ReleaseShared();
            QueueLock.ReleaseExclusive();

            foreach (Navigator nav in navs)
            {
                
                try
                {
                    if (MyApiWrapper.TryGetAiLock("autopilot running update"))
                    {
                        nav.update();
                    }
                    else
                    {
                        return;
                    }
                }
                catch (Exception e)
                {
                    MissionComponent.Speak("AutoPilotUpdater", "Error encountered:\n" + e);
                }
                finally
                {
                    MyApiWrapper.ReleaseAiLock();
                }
                
            }
            

        }
    }
}
