﻿using System;
using System.Collections;
using Sandbox.Common.ObjectBuilders;
using Sandbox.ModAPI;
using Sandbox.ModAPI.Interfaces;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Sandbox.Definitions;
using VRageMath;

namespace Scripts.KSWH
{
    /// <summary>
    /// Class to execute electronic counter measures
    /// </summary>
    class EcmActions
    {
        private string[] exceptions;
        private bool empTriggered = false;
        private int empTicker, empCode;
        private bool working = false;
        private bool shuttingdownPower = true;
        private string[] mOreNames;

        public enum Size
        {
            Small,
            Large
        }

        /// <summary>
        /// Default constructor
        /// </summary>
        public EcmActions()
        {
            MyDefinitionManager.Static.GetOreTypeNames(out mOreNames);
        }

        /// <summary>
        /// Send in a storm of old scrapmetal to hit the target
        /// </summary>
        /// <param name="size">size of the scrap grids</param>
        /// <param name="hitLocation"> location of impact</param>
        public void ScrapStorm(Size size, Vector3D hitLocation)
        {
            MissionComponent.AlarmHandler.EcmWarning();
            if (MissionComponent.writeDebug)
            {
                MissionComponent.Speak("ECM", "Activating magnetic storm");
            }
            Vector3D spawnLocation = new Vector3D(hitLocation.X + 20, hitLocation.Y + 30, hitLocation.Z);
            Random r = new Random();
            double x = (spawnLocation.X - hitLocation.X) * (spawnLocation.X < hitLocation.X ? 10 : -10);
            double y = (spawnLocation.Y - hitLocation.Y) * (spawnLocation.Y < hitLocation.Y ? 10 : -10);
            double z = (spawnLocation.Z - hitLocation.Z) * (spawnLocation.Z < hitLocation.Z ? 10 : -10);
            

            switch (size)
            {
                case Size.Small:
                    //spawn in 5 small wrecks on its way to hit target
                    for (int i = 0; i < 5; i++)
                    {
                        ThreadWrapper.OrderSpawn(
                            StaticInformation.SmallWreckages[r.Next(StaticInformation.SmallWreckages.Length - 1)],
                            new Vector3D(spawnLocation.X + (i*5), spawnLocation.Y + (i * 20), spawnLocation.Z + (i*5)),
                            new Vector3D(x - (i*5), y + (i * 20), z - (i*5)));
                    }
                    break;
                case Size.Large:
                    //spawn in 10 small wrecks on its way to hit target
                    for (int i = 0; i < 10; i++)
                    {
                        ThreadWrapper.OrderSpawn(
                            StaticInformation.SmallWreckages[r.Next(StaticInformation.SmallWreckages.Length - 1)],
                            new Vector3D(spawnLocation.X + (i * 5), spawnLocation.Y + (i * 20), spawnLocation.Z + (i * 5)),
                            new Vector3D(x - (i * 5), y + (i * 20), z - (i * 5)));
                    }
                    //spawn in 3 large wrecks
                    for (int i = 9; i < 13; i++)
                    {
                        x = ((spawnLocation.X + (i * 5)) - hitLocation.X) * (spawnLocation.X < hitLocation.X ? 10 : -10);
                        y = ((spawnLocation.Y + (i * 40)) - hitLocation.Y) * (spawnLocation.Y < hitLocation.Y ? 10 : -10);
                        z = ((spawnLocation.Z + (i * 5)) * (spawnLocation.Z < hitLocation.Z ? 10 : -10));
                        ThreadWrapper.OrderSpawn(
                            StaticInformation.LargeWreckages[r.Next(StaticInformation.LargeWreckages.Length - 1)],
                            new Vector3D(spawnLocation.X + (i * 10), spawnLocation.Y + (i * 40), spawnLocation.Z + (i * 10)),
                            new Vector3D(x, y, z));
                    }
                    break;
            }
        }

        /// <summary>
        /// Send in a meteor storm
        /// </summary>
        /// <param name="hitLocation">Location of impact</param>
        /// <param name="startLocation">Location where the meteor coames from</param>
        /// <param name="shooter">The entity that is firing the strike</param>
        /// <param name="size">Size of the attack (small or big)</param>
        public void LaunchMetoer(Vector3D hitLocation, Vector3D startLocation, MatrixD shooter, Size size)
        {
            MissionComponent.AlarmHandler.EcmWarning();
            if (MissionComponent.writeDebug)
            {
                MissionComponent.Speak("ECM", "Launching astroid strike");
            }
            Vector3D position = new Vector3D(startLocation.X + 10, startLocation.Y - 20, startLocation.Z);
            Random r = new Random();
            
            var meteorBuilder = new MyObjectBuilder_Meteor()
            {
                Item = new MyObjectBuilder_InventoryItem()
                {
                    Amount = (size == Size.Small ? 10000 : 100000),
                    Content = new MyObjectBuilder_Ore() { SubtypeName = mOreNames[r.Next(mOreNames.Length - 1)] }
                },
                PersistentFlags = MyPersistentEntityFlags2.InScene, // Very important
                PositionAndOrientation = new MyPositionAndOrientation()
                {
                    Position = position,
                    Forward = (Vector3)shooter.Forward,
                    Up = (Vector3)shooter.Up,
                },
                LinearVelocity = shooter.Forward * 300,
                Integrity = 100,
            };
            var tempList = new List<MyObjectBuilder_EntityBase> { meteorBuilder };
            MyAPIGateway.Entities.RemapObjectBuilderCollection(tempList);
            tempList.ForEach(grid => MyAPIGateway.Entities.CreateFromObjectBuilderAndAdd(grid));
            MyAPIGateway.Multiplayer.SendEntitiesCreated(tempList);
        }

        /// <summary>
        /// Send out a blast that will disable all power of nearby ships
        /// Drain all nearby bateries to 0% power
        /// push every object away
        /// </summary>
        /// <param name="range">range in meters that the blast should go</param>
        public void EmpBlast(BoundingSphereD range)
        {
            MissionComponent.AlarmHandler.EcmWarning();
            if (MissionComponent.writeDebug)
            {
                MissionComponent.Speak("ECM", "Pulsing EMP blast");
            }
            //todo: implement
            MissionComponent.Speak("ECM", "EMP blast not implemented");
        }

    }
}
