﻿using Sandbox.Common.ObjectBuilders;
using Sandbox.ModAPI;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ParallelTasks;
using Sandbox.Common;
using VRage;
using VRageMath;

namespace Scripts.KSWH
{
    /// <summary>
    /// simple data class to keep track of the host data
    /// </summary>
    class HostData
    {
        public long mPlayerId { private set; get; }
        public string PlayerFactionTag { private set; get; }
        public long enemy { private set; get; }
        public bool ready { private set; get; } //bool if mod is ready to run
        public bool run { set; get; } //bool if mod is running
        public bool isSave { private set; get; }
        public bool saveRun { private set; get; }
        private readonly string mGameName;
        public int AiResources { set; get; }
        public int AiHookSlots { set; get; }
        public int ActiveHooks { set; get; }
        private int ShipId = 0;
        private FastResourceLock idLock = new FastResourceLock();
        private string overFlowData;
        private FastResourceLock writeloLock = new FastResourceLock();
        public int playerResources { private set; get; }

        private int cAiResources;
        private bool cRun;
        private int cAiHookCount;
        private string cTag;
        private int cplayerResources;

        /// <summary>
        /// default constructor
        /// </summary>
        /// <param name="playerId">steamid of the player</param>
        public HostData(long playerId)
        {
            mPlayerId = playerId;
            //mGameName = MyAPIGateway.Session.Name + "_" + mPlayerId;
            mGameName = "Sector15v2";
            enemy = 384; //random value
            if (MyAPIGateway.Utilities.FileExistsInLocalStorage(mGameName, typeof(HostData)))
            {
                MissionComponent.Speak("Wartank", "Loading game state, please wait...");
                LoadSettings();
                isSave = true;
                List<IMyPlayer> player = new List<IMyPlayer>();
                MyAPIGateway.Players.GetPlayers(player);
            }
            else
            {
                MissionComponent.Notification("New game started", 30000, MyFontEnum.Blue);
                isSave = false;
                AiHookSlots = 5; //default
            }

        }

        /// <summary>
        /// set the faction hook for the mod to run on
        /// </summary>
        /// <param name="tag"></param>
        public void SetFactionTag(string tag)
        {
            PlayerFactionTag = tag;
            ready = true;
            //WriteSettings();
        }

        /// <summary>
        /// Get a unique id for a ship
        /// </summary>
        /// <returns>a unique id of a ship</returns>
        public int CreateshipId()
        {
            idLock.AcquireExclusive();
            var ret = ShipId;
            ShipId++;
            idLock.ReleaseExclusive();
            return ret;
        }

        /// <summary>
        /// start the actual mod
        /// </summary>
        public void Start()
        {
            if (AiHookSlots < 1)
            {
                MyAPIGateway.Utilities.ShowMissionScreen("WARNING", "Configuration: ", "questionable",
                         "The system noticed that you allowed a max ammount of enemies equal to " + AiHookSlots + " This will infortunally mean that no enemy ships will spawn (where is the challange in that?)\n\n" +
                         "The system has restored the default setting of a maximum of 5 hostiles. To change this type /maxAi [NUMBER]. Please keep in mind that high numbers may cause performance issues on" +
                         "Computers with a low core speed.");
                AiHookSlots = 5;
                ForceSave();

            }
            run = true;
            WriteSettings();
        }

        /// <summary>
        /// check if the player can pay for the request.
        /// </summary>
        /// <param name="amount">amount the player needs to pay</param>
        /// <returns>true if he could pay it and the money has been taken. False if he did not have enough money</returns>
        public bool PlayerPay(int amount)
        {
            if (playerResources < amount) return false;
            playerResources -= amount;
            return true;
        }

        /// <summary>
        /// Reward the player a bounty
        /// </summary>
        /// <param name="amount">amount of money the player gains</param>
        public void PlayerGain(int amount)
        {
            if (run)
            {
                MissionComponent.hud.AddMessage((int)Hud.MessagePriority.Other, Hud.MessageLead + "//PAYMENT\nBounty of " + amount + " awarded");
            }
            playerResources += amount;
        }

        /// <summary>
        /// stop the game
        /// </summary>
        public void Stop()
        {
            run = false;
            WriteSettings();
        }

        public void ForceSave()
        {
            WriteSettings();
        }

        /// <summary>
        /// make the ai pay for a ship
        /// </summary>
        /// <param name="ammount">ammount of money it needs to pay</param>
        /// <returns>bool if the ai could afford it</returns>
        public bool AiPay(int ammount)
        {
            if (AiResources - ammount > 0)
            {
                AiResources -= ammount;
                return true;
            }
            return false;
        }

        /// <summary>
        /// Check if the ai can pay for something
        /// </summary>
        /// <param name="ammount">ammount of money it would have to pay</param>
        /// <returns></returns>
        public bool AiHasMoney(int ammount)
        {
            return AiResources >= ammount;
        }

        /// <summary>
        /// write settings file to screen
        /// </summary>
        public void DisplaySettingsFile()
        {
            var reader = MyAPIGateway.Utilities.ReadFileInLocalStorage(mGameName, typeof(HostData));
            bool keepReading = true;
            do
            {
                string line = reader.ReadLine();
                if (line == null)
                {
                    keepReading = false;
                }
                else
                {
                    MissionComponent.Speak("file line:", line);
                }
            } while (keepReading);
            reader.Close();
        }

        /// <summary>
        /// write game data to file (save)
        /// </summary>
        private void WriteSettings()
        {
           if (writeloLock.TryAcquireExclusive())
            {
                cAiHookCount = AiHookSlots;
                cAiResources = AiResources;
                cTag = PlayerFactionTag;
                cRun = run;
                cplayerResources = playerResources;
                MyAPIGateway.Parallel.StartBackground(WriteSettingsAsync);
            }
            

        }

        private void WriteSettingsAsync()
        {
            if (MissionComponent.writeDebug)
            {
                MissionComponent.Speak("Optionhandler", "Writing data");
            }
            
            string gameName = MyAPIGateway.Session.Name;

            try
            {
                var writer = MyAPIGateway.Utilities.WriteFileInLocalStorage(mGameName, typeof (HostData));

                StringBuilder sb = new StringBuilder();
                sb.Append(gameName);
                sb.Append("#");
                sb.Append(cRun);
                sb.Append('|');
                sb.Append(cTag);
                sb.Append('|');
                sb.Append(cAiResources);
                sb.Append('|');
                sb.Append(cAiHookCount);
                sb.Append('|');
                sb.Append(cplayerResources);
                writer.WriteLine(sb.ToString());
                writer.WriteLine(overFlowData);
                writer.Close();
            }
            catch (Exception e)
            {
                MissionComponent.Speak("Wartank", "****WARNING! saving failed!****");
                if (MissionComponent.writeDebug)
                {
                    MissionComponent.Speak("system", e.ToString());
                }
            }
            finally
            {
                writeloLock.ReleaseExclusive();
                if (MissionComponent.writeDebug)
                {
                    MissionComponent.Speak("Optionhandler", "Writing data complete");
                }
            }
        }


        private void LoadSettings()
        {
            if (!MyAPIGateway.Utilities.FileExistsInLocalStorage(mGameName, typeof(HostData)))
            {
                return; //protection against file not found
            }

            var reader = MyAPIGateway.Utilities.ReadFileInLocalStorage(mGameName, typeof (HostData));
            bool keepReading = true;
            string session = MyAPIGateway.Session.Name;
            StringBuilder sb = new StringBuilder();
            while (keepReading)
            {
                string line = reader.ReadLine();
                if (line == null)
                {
                    break;
                }
                string[] splitheader = line.Split('#');
                if (splitheader.Length != 2)
                {
                    MissionComponent.Speak("WARNING: ", "corruption in file, splitheader: " + splitheader.Length + "\n" + line);
                    continue;
                }
                if (splitheader[0] == session)
                {
                    string[] values = splitheader[1].Split('|');
                    if (values.Length != 5)
                    {
                        MissionComponent.Speak("WARNING:", "Failed to load save file, the file appears to be corrupt, new session created. errcode: " + values.Length);
                        return;
                    }
                    //run = !values[0].Contains("False");
                    PlayerFactionTag = values[1];
                    AiResources = Convert.ToInt32(values[2]);
                    AiHookSlots = Convert.ToInt32(values[3]);
                    playerResources = Convert.ToInt32(values[4]);

                }
                else
                {
                    sb.Append(line);
                }
                
            }
            overFlowData = sb.ToString();
        }


        /// <summary>
        /// get game data from file (load)
        /// </summary>
        [Obsolete("This is the load function for the save files v1", true)]
        private void LoadSettings_Old()
        {
            if (!MyAPIGateway.Utilities.FileExistsInLocalStorage(mGameName, typeof(HostData)))
            {
                return; //protection against file not found
            }

            var reader = MyAPIGateway.Utilities.ReadFileInLocalStorage(mGameName, typeof(HostData));
            bool keepReading = true;
            do
            {
                string line = reader.ReadLine();
                if (!string.IsNullOrEmpty(line))
                {
                    if (line.Contains("[run]") && MissionComponent.hasFirewall)
                    {
                        ready = !line.Contains("False");
                        if (ready && MissionComponent.FindFirewall() == null)
                        {
                            ready = false;
                        }

                    } else if (line.Contains("[tag]"))
                    {
                        PlayerFactionTag = line.Replace("[tag]", "").Trim();
                    }
                    else if (line.Contains("[ar]"))
                    {
                        AiResources = Convert.ToInt32(line.Replace("[ar] ", ""));
                    }
                    else if (line.Contains("[cs"))
                    {
                        AiHookSlots = Convert.ToInt32(line.Replace("[cs] ", ""));
                    }
                }
                else
                {
                    keepReading = false;
                }
                
            } while (keepReading);
            reader.Close();
            if (AiHookSlots < 5)
            {
                AiHookSlots = 5;
            }
            saveRun = false;
        }

    }
}
