﻿using System;
using System.Collections.Generic;
using System.Linq;
using Sandbox.Common.ObjectBuilders;
using Sandbox.Definitions;
using Sandbox.ModAPI;
using Sandbox.ModAPI.Interfaces;
using VRageMath;

namespace Scripts.KSWH
{
    class Spawner
    {
        /// <summary>
        /// Spawn in a new entity to the world and sync that up with the clients
        /// </summary>
        /// <param name="prefabName">name of the prefab to load</param>
        /// <param name="location">Location where the prefab syou thould be spawned</param>
        /// <returns>a list of entities that was created. index 0 is ALWAYS the main entity. Rest are thigns connected to rotors and pistons</returns>
        public static List<IMyEntity> Spawn(string prefabName, Vector3D location)
        {
            ThreadWrapper.GetSpawnLock();
            //looks for the definition of the ship
            var prefab = MyDefinitionManager.Static.GetPrefabDefinition(prefabName);

            if (prefab == null)
            {
                MissionComponent.Speak("Error", "Failed to spawn " + prefabName + " prefab not found (file missing?)");
                ThreadWrapper.ReleaseSpawnLock();
                return null;
            }
            if (prefab.CubeGrids == null)
            {
                MyDefinitionManager.Static.ReloadPrefabsFromFile(prefab.PrefabPath);
                prefab = MyDefinitionManager.Static.GetPrefabDefinition(prefabName);
            }

            if (MissionComponent.writeDebug)
            {
                MissionComponent.Speak("system", "Prefab has " + prefab.CubeGrids.Count() + " cubegrids");
            }
            
            List<MyObjectBuilder_EntityBase> tempList = new List<MyObjectBuilder_EntityBase>();
            List<IMyEntity> output = new List<IMyEntity>();
            foreach (var grid in prefab.CubeGrids)
            {
                tempList.Add(grid);
                
                //give the grid containing the ship a new position
                grid.PositionAndOrientation = new MyPositionAndOrientation(location, Vector3D.Forward, Vector3D.Up);
                Random r = new Random(DateTime.Now.Millisecond);
                grid.EntityId = r.Next(); //don't forget to give it a new id or it will reuse the old (aka despawn/respawn)
                MyAPIGateway.Entities.RemapObjectBuilder(grid);
                //Create the object and add it to the world.
                output.Add(MyAPIGateway.Entities.CreateFromObjectBuilderAndAdd(grid));
                if (MyAPIGateway.Multiplayer.IsServer)
                {
                    MyAPIGateway.Multiplayer.SendEntitiesCreated(tempList);
                }
                //important ! every entity has a unique ID. If you try to add your prefab twice, they would have the same ID so that wont work
                //so you use RemapObjectBuilder to generate new IDs for the ship and other entities onboard.
            }
            ThreadWrapper.ReleaseSpawnLock();
            return output;
        }

        /// <summary>
        /// spawn in a prefab and given it a certain speed
        /// This method respects the rules of thread locking and is save to call
        /// </summary>
        /// <param name="prefabName">name of prefab to spawn</param>
        /// <param name="location">location where the prefab should popup</param>
        /// <param name="speed">speed how quick the prefab should fly in each direction.</param>
        public static void SpawnWithSpeed(string prefabName, Vector3D location, Vector3D speed)
        {
            ThreadWrapper.GetSpawnLock();
            //looks for the definition of the ship
            var prefab = MyDefinitionManager.Static.GetPrefabDefinition(prefabName);

            if (prefab == null)
            {
                MissionComponent.Speak("Error", "Failed to spawn " + prefabName + " prefab not found (file missing?)");
                ThreadWrapper.ReleaseSpawnLock();
                return;
            }
            if (prefab.CubeGrids == null)
            {
                MyDefinitionManager.Static.ReloadPrefabsFromFile(prefab.PrefabPath);
                prefab = MyDefinitionManager.Static.GetPrefabDefinition(prefabName);
            }

            if (MissionComponent.writeDebug)
            {
                MissionComponent.Speak("system", "Prefab has " + prefab.CubeGrids.Count() + " cubegrids");
            }

            List<MyObjectBuilder_EntityBase> tempList = new List<MyObjectBuilder_EntityBase>();
            foreach (var grid in prefab.CubeGrids)
            {
                tempList.Add(grid);

                //give the grid containing the ship a new position
                grid.PositionAndOrientation = new MyPositionAndOrientation(location, Vector3D.Forward, Vector3D.Up);
                Random r = new Random(DateTime.Now.Millisecond);
                grid.EntityId = r.Next(); //don't forget to give it a new id or it will reuse the old (aka despawn/respawn)
                MyAPIGateway.Entities.RemapObjectBuilder(grid);
                //Create the object and add it to the world.
                var item = MyAPIGateway.Entities.CreateFromObjectBuilderAndAdd(grid);
                if (MyAPIGateway.Multiplayer.IsServer)
                {
                    MyAPIGateway.Multiplayer.SendEntitiesCreated(tempList);
                }
                item.Physics.LinearVelocity = speed;
                //important ! every entity has a unique ID. If you try to add your prefab twice, they would have the same ID so that wont work
                //so you use RemapObjectBuilder to generate new IDs for the ship and other entities onboard.
            }
            ThreadWrapper.ReleaseSpawnLock();
        }

        /// <summary>
        /// WARNING: do not use other then debug. this method is NOT thread save.
        /// </summary>
        /// <param name="prefabName"></param>
        /// <param name="location"></param>
        /// <returns></returns>
        public static List<IMyEntity> UnsaveSpawn(string prefabName, Vector3D location)
        {
            
            //looks for the definition of the ship
            var prefab = MyDefinitionManager.Static.GetPrefabDefinition(prefabName);

            if (prefab == null)
            {
                MissionComponent.Speak("Error", "Failed to spawn " + prefabName + " prefab not found (file missing?)");
                ThreadWrapper.ReleaseSpawnLock();
                return null;
            }

            if (prefab.CubeGrids == null)
            {
                MyDefinitionManager.Static.ReloadPrefabsFromFile(prefab.PrefabPath);
                prefab = MyDefinitionManager.Static.GetPrefabDefinition(prefabName);
            }

            if (MissionComponent.writeDebug)
            {
                MissionComponent.Speak("system", "Prefab has " + prefab.CubeGrids.Count() + " cubegrids");
            }

            List<MyObjectBuilder_EntityBase> tempList = new List<MyObjectBuilder_EntityBase>();
            List<IMyEntity> output = new List<IMyEntity>();
            MissionComponent.Speak("spawner", "templist created");
            foreach (var grid in prefab.CubeGrids)
            {
                
                tempList.Add(grid);
                MissionComponent.Speak("spawner", "added item to list");
                //give the grid containing the ship a new position
                grid.PositionAndOrientation = new MyPositionAndOrientation(location, Vector3D.Forward, Vector3D.Up);
                MissionComponent.Speak("spawner", "rotation set");
                Random r = new Random(DateTime.Now.Millisecond);
                grid.EntityId = r.Next(); //don't forget to give it a new id or it will reuse the old (aka despawn/respawn)
                MissionComponent.Speak("spawner", "created id " + grid.EntityId);
                MyAPIGateway.Entities.RemapObjectBuilder(grid);
                MissionComponent.Speak("spawner", "remap complete");
                //Create the object and add it to the world.
                output.Add(MyAPIGateway.Entities.CreateFromObjectBuilderAndAdd(grid));
                
                
                //important ! every entity has a unique ID. If you try to add your prefab twice, they would have the same ID so that wont work
                //so you use RemapObjectBuilder to generate new IDs for the ship and other entities onboard.
            }

            if (MyAPIGateway.Multiplayer.IsServer)
            {
                MyAPIGateway.Multiplayer.SendEntitiesCreated(tempList);
            }

            return output;
        }
        
    }
}
