﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Sandbox.Common.ObjectBuilders;
using Sandbox.Definitions;
using Sandbox.ModAPI;
using Sandbox.ModAPI.Interfaces;
using VRageMath;

namespace Scripts.KSWH
{
    public abstract class Station : Controlable
    {
        public enum StationType
        {
            Frigate,
            Destroyer,
            SmallLauncher,
            SmallLogistic,
            Cruiser,
            Battleship,
            Ecm,
            Capital,
            LargeLauncher,
            LargeLogistics
        }


        
        public StationType Type { protected set; get; }
        public bool Online { private set; get; }
        protected int Income;
        public bool Usefull = true;
        protected bool mSelfdestructing;
        protected int CurrentProductionTime;
        protected DateTime ProductionStart;
        protected string ConstructionPrefabName;
        public bool IsProducing { protected set; get; }
        protected int XOffset, YOffset, ZOffset;
        protected bool CalculatedOffset;

        public void Update()
        {
            ManageResources();
            if (IsProducing)
            {
                ManageProduction();
            }
            Usefull = IsValid();
        }

        

        /// <summary>
        /// give the AI rescources
        /// </summary>
        protected virtual void ManageResources()
        {
            MissionComponent.host.AiResources += Income;
        }

        protected virtual void ManageProduction()
        {
            TimeSpan span = (DateTime.Now - ProductionStart);
            if (span.TotalMinutes + MissionComponent.AiSpeedModifier >= CurrentProductionTime)
            {
                if (!CalculatedOffset)
                {
                    var firewall = MissionComponent.firewall;
                    if (firewall == null)
                    {
                        //no firewall found, go to default, odd that this happened
                        XOffset = 100;
                        YOffset = 0;
                        ZOffset = 0;
                        CalculatedOffset = true;
                    }
                    else
                    {
                        var firewallPos = firewall.CubeGrid.GetPosition();
                        var thispos = Grid.GetPosition();
                        if (firewallPos.X > thispos.X)
                        {
                            XOffset = -100;
                        }
                        else
                        {
                            XOffset = 100;
                        }
                        if (firewallPos.Y > thispos.Y)
                        {
                            YOffset = -100;
                        }
                        else
                        {
                            YOffset = 100;
                        }
                        if (firewallPos.Z > thispos.Z)
                        {
                            ZOffset = -100;
                        }
                        else
                        {
                            ZOffset = 100;
                        }
                        CalculatedOffset = true;
                    }
                }


                Vector3D currentPosition = Grid.GetPosition();

                if (MissionComponent.writeDebug)
                {
                    StringBuilder sb = new StringBuilder();
                    sb.Append("Deploying ship on ");
                    sb.Append(currentPosition.X + XOffset);
                    sb.Append(", ");
                    sb.Append(currentPosition.Y + YOffset);
                    sb.Append(", ");
                    sb.Append(currentPosition.Z + YOffset);
                    MissionComponent.Speak("Station", sb.ToString());
                }
                ThreadWrapper.OrderSpawn(ConstructionPrefabName,
                    new Vector3D(currentPosition.X + XOffset, currentPosition.Y + YOffset, currentPosition.Z + ZOffset), Type, this);
            }
        }

        /// <summary>
        /// clear the production slot when the ship has spawned
        /// </summary>
        public void FreeProductionSlot()
        {
            IsProducing = false;
            GoOffline();
        }

        /// <summary>
        /// Begin a production order
        /// </summary>
        /// <param name="duration">time it has to take (in  minutes)</param>
        /// <param name="constructionPrefabName"> name of the prefab you want to construct</param>
        /// <returns>bool if the operation started</returns>
        protected bool StartProduction(int duration, string constructionPrefabName)
        {
            if (IsProducing)
            {
                return false; //stations can only have 1 production slot
            }
            GoOnline();
            ConstructionPrefabName = constructionPrefabName;
            CurrentProductionTime = duration;
            ProductionStart = DateTime.Now;
            IsProducing = true;
            return true;
        }


        /// <summary>
        /// station is in production mode
        /// </summary>
        private void GoOnline()
        {
            Online = true;
            SwitchStationActivity();
        }

        /// <summary>
        /// Station is no longer in production mode
        /// </summary>
        protected void GoOffline()
        {
            Online = false;
            SwitchStationActivity();
        }

        /// <summary>
        /// Fucntion to check if the station is still valid for the AI
        /// </summary>
        /// <returns>bool if station is valid</returns>
        protected bool IsValid()
        {
            if (GetCockpit() != null)
            {
                return true;
            }
            else
            {
                SelfDestruct();
                MissionComponent.host.PlayerGain(GetBounty());
                return false;
            }
        }

        /// <summary>
        /// Destroy the station
        /// </summary>
        private void SelfDestruct()
        {

            if (mSelfdestructing)
            {
                return;
            }

            mSelfdestructing = true;

            if (MissionComponent.writeDebug)
            {
                MissionComponent.Speak("station", "Self destruct activated");
            }
            try
            {
                var lst = new List<IMySlimBlock>();
                MyDefinitionId nukeDef = new MyDefinitionId(typeof (MyObjectBuilder_Warhead));
                Grid.GetBlocks(lst, x => x.FatBlock != null && x.FatBlock.BlockDefinition.TypeId.Equals(nukeDef.TypeId));
                if (lst.Count <= 0)
                {
                    return; //no nukes on grid
                }
                var nuke = lst[0];
                var actions = new List<ITerminalAction>();
                MyAPIGateway.TerminalActionsHelper.GetActions(nuke.FatBlock.GetType(), actions,
                    x => x.Id.Contains("Start"));
                if (actions.Count == 0)
                {
                    return; //no action found
                }
                var action = actions.FirstOrDefault();
                foreach (IMySlimBlock block in lst)
                {
                    action.Apply(block.FatBlock);
                }
            }
            catch (Exception)
            {
                //failsave for if all warheads are destoryed, do nothing
            }
        }

        private void SwitchStationActivity()
        {
            try
            {
                if (MissionComponent.writeDebug)
                {
                    MissionComponent.Speak("Station", "Switching online state to " + Online);
                }
                var lst = new List<IMySlimBlock>();
                var name = Online ? "!start" : "!stop";
                MyDefinitionId timerDef = new MyDefinitionId(typeof (MyObjectBuilder_TimerBlock));
                Grid.GetBlocks(lst,
                    x =>
                        x.FatBlock != null && x.FatBlock.BlockDefinition.TypeId.Equals(timerDef.TypeId) &&
                        x.FatBlock.DisplayNameText.ToLower().Equals(name));
                if (lst.Count < 0)
                {
                    if (MissionComponent.writeDebug)
                    {
                        MissionComponent.Speak("Station", "Could not set activity of station, no timer block found");
                    }
                    return; //station damaged and timer block is gone....
                }
                var timer = lst[0];
                var actions = new List<ITerminalAction>();
                MyAPIGateway.TerminalActionsHelper.GetActions(timer.FatBlock.GetType(), actions,
                    x => x.Id.Contains("Start"));
                if (actions.Count == 0)
                {
                    if (MissionComponent.writeDebug)
                    {
                        MissionComponent.Speak("Station", "No timer action found");
                    }
                    return; //no action found
                }
                var action = actions[0];
                foreach (IMySlimBlock timert in lst)
                {
                    action.Apply(timert.FatBlock);
                }
            }
            catch (Exception)
            {
                //failsave for if all timers are destoryed
            }
        }

       /// <summary>
       /// Function that craetes a new location for a station to spawn
       /// </summary>
       /// <param name="minimumDisantace">minimum dinstace from the firewall</param>
       /// <returns>A vector 3d with a ideal location for the station</returns>
        protected Vector3D GenerateLocation(int minimumDisantace)
        {
            Random r = new Random();
            int x = (r.Next(10) + minimumDisantace) * 1000 + MissionComponent.firewall.Position.X * (r.Next(1) == 1 ? 1 : -1); //if (rdn == 1) { return 1} else {return -1}
            int y = (r.Next(10) + minimumDisantace) * 1000 + MissionComponent.firewall.Position.Y * (r.Next(1) == 1 ? 1 : -1);
            int z = (r.Next(10) + minimumDisantace) * 1000 + MissionComponent.firewall.Position.Z * (r.Next(1) == 1 ? 1 : -1);
            return new Vector3D(x, y, z);
        }

        
    }
}
