﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Scripts.KSWH
{
    /// <summary>
    /// Class that will manage the incursion stations
    /// </summary>
    internal class StationManager
    {

        public List<Station> Stations { private set; get; }
        private DateTime LastPingTime;
        private ProductionOrder order;
        private bool IsBuildingStation;
        private int baseIncome = 50;
        public bool EcmReady { private set; get; } //todo: implement em setter
        public int IncursionState { private set; get; }

        /// <summary>
        /// Small struct that can hold a station production order
        /// </summary>
        private struct ProductionOrder
        {
            public int StationProductionTime;
            public DateTime ProductionStart;
            public Station.StationType Type;

            /// <summary>
            /// Default constructor
            /// </summary>
            /// <param name="time">Time it will take the build</param>
            /// <param name="type">Type of station bieng build</param>
            public ProductionOrder(int time, Station.StationType type)
            {
                StationProductionTime = time;
                Type = type;
                ProductionStart = DateTime.Now;
            }

            /// <summary>
            /// Send the order into the game world
            /// </summary>
            /// <returns>the station object that was created</returns>
            public Station DeployOrder()
            {
                MissionComponent.AlarmHandler.NewSig();
                switch (Type)
                {
                    case Station.StationType.Frigate:
                        return new FrigateStation();
                    case Station.StationType.Destroyer:
                        return new DestroyerStation();
                    case Station.StationType.SmallLauncher:
                        return new SmallLaunchPlatform();
                    case Station.StationType.SmallLogistic:
                        return new SmallFactoryStation();
                    case Station.StationType.Cruiser:
                        return new CruiserStation();
                    case Station.StationType.Battleship:
                        return new BattleshipStation();
                    case Station.StationType.Ecm:
                        return new EcmStation();
                    case Station.StationType.Capital:
                        return new CapitalStation();
                    case Station.StationType.LargeLauncher:
                        return new LargeLaunchPlatform();
                    case Station.StationType.LargeLogistics:
                        return new LargeFactoryStation();
                }
                return null;
            }
        }

        /// <summary>
        /// Constructor
        /// </summary>
        public StationManager()
        {
            Stations = new List<Station>();
            //todo: load station from world
            LastPingTime = DateTime.Now;
        }

        public void SetStations(List<Station> stations)
        {
            Stations = stations;
        }

        /// <summary>
        /// Update all known stations
        /// </summary>
        public void Update()
        {
            UpdateTask();
            //MyAPIGateway.Parallel.StartBackground(UpdateTask);
        }

        private void UpdateTask()
        {
            TimeSpan span = (DateTime.Now - LastPingTime);
            if (span.Minutes >= 1)
            {
                List<Station> uselessStations = new List<Station>();

                foreach (Station s in Stations)
                {
                    s.Update();
                    if (!s.Usefull)
                    {
                        uselessStations.Add(s);
                    }
                }


                foreach (var station in uselessStations)
                {
                    if (MissionComponent.writeDebug)
                    {
                        MissionComponent.Speak("StationManager", "Removing Station");
                    }
                    Stations.Remove(station);
                }
                LastPingTime = DateTime.Now;

                MissionComponent.host.AiResources += baseIncome;
                if (baseIncome < 300)
                {
                    baseIncome++;
                }

                PlotNextMove(); //plot next move before finalizing station production to make the game a little easier

                if (IsBuildingStation)
                {
                    TimeSpan buildSpan = (DateTime.Now - order.ProductionStart);
                    if (buildSpan.Minutes + MissionComponent.AiSpeedModifier >= order.StationProductionTime)
                    {
                        Stations.Add(order.DeployOrder());
                        IsBuildingStation = false;
                        if (MissionComponent.writeDebug)
                        {
                            MissionComponent.Speak("StationManager", "Completed station");
                        }
                    }
                    else if (MissionComponent.writeDebug)
                    {
                        MissionComponent.Speak("StationManager",
                            "Constructing station " + buildSpan.Minutes + "/" + order.StationProductionTime);
                    }
                }
                if (MissionComponent.writeDebug)
                {
                    StringBuilder sb = new StringBuilder();
                    sb.Append("finished turn, ");
                    sb.Append(MissionComponent.host.AiResources);
                    sb.Append("$ stations: ");
                    sb.Append(Stations.Count());
                    sb.Append(". Isbuilding new station: ");
                    sb.Append(IsBuildingStation);
                    MissionComponent.Speak("StationManager", sb.ToString());
                }
                MissionComponent.host.ForceSave(); //write resources to file
            }
            //MissionComponent.sManager.StationSimulationRunning = false;
        }

        /// <summary>
        /// Make the AI decide what build orders to start
        /// </summary>
        private void PlotNextMove()
        {
            if (MissionComponent.AiLivePing)
            {
                MissionComponent.Speak("ai", "counting stations");
            }
            IncursionState = 0;
            var frigateStations = Stations.Where(x => x.Type == Station.StationType.Frigate);
            var cruiserStation = Stations.Where(x => x.Type == Station.StationType.Cruiser);
            var battleshipStation = Stations.Where(x => x.Type == Station.StationType.Battleship);
            var smallLogisticStations = Stations.Where(x => x.Type == Station.StationType.SmallLogistic);
            var smallLauncherStation = Stations.Where(x => x.Type == Station.StationType.SmallLauncher);
            var destroyerStation = Stations.Where(x => x.Type == Station.StationType.Destroyer);
            var largeLauncherStation = Stations.Where(x => x.Type == Station.StationType.LargeLauncher);
            var largeLogisticsStation = Stations.Where(x => x.Type == Station.StationType.LargeLogistics);
            var ecmStation = Stations.FirstOrDefault(x => x.Type == Station.StationType.Ecm);
            var capitalStations = Stations.Where(x => x.Type == Station.StationType.Capital);
            int fStationCount = frigateStations.Count();
            int cStationCount = cruiserStation.Count();
            int bStationCount = battleshipStation.Count();
            int sLStationCount = smallLogisticStations.Count() + smallLauncherStation.Count();
            int lMStationCount = largeLauncherStation.Count();
            int LLStationCount = largeLogisticsStation.Count() + (ecmStation == null ? 0 : 1);
            int csStationCount = capitalStations.Count();
            int dStationCount = destroyerStation.Count();

            if (MissionComponent.AiLivePing)
            {
                MissionComponent.Speak("ai", "getting incursion level");
            }
            IncursionState = DeterminIncursionState(dStationCount, sLStationCount,
                        cStationCount, lMStationCount, ecmStation, LLStationCount,
                        bStationCount, cStationCount);

            if (!IsBuildingStation)
            {
                //building slot ready
                /*if (fStationCount == 0)
                {
                    if (MissionComponent.writeDebug)
                    {
                        MissionComponent.Speak("StationManager", "Shopping for new station");
                    }
                    //state 1 is incomplete, highest priority
                    if (MissionComponent.host.AiPay(FrigateStation.BuildCosts))
                    {
                        StartStationProduction(Station.StationType.Frigate);
                    }
                    else
                    {
                        return; //no need to do a thing, state 1 MUST be completed first
                    }
                }*/
                if (IncursionState == 0)
                {
                    IncursionState = 1;
                }
                if (MissionComponent.AiLivePing)
                {
                    MissionComponent.Speak("ai", "selecting station to build");
                }
                if (IncursionState > 0 && IncursionState < 7)
                {
                    //we are not at capital levels yet. either push to attack or expand

                    if (fStationCount + 1 < IncursionState)
                    {
                        if (MissionComponent.host.AiPay(FrigateStation.BuildCosts))
                        {
                            if (MissionComponent.AiLivePing)
                            {
                                MissionComponent.Speak("Station manager", "Building Frigate Station");
                            }
                            StartStationProduction(Station.StationType.Frigate);

                        }
                    }
                    else if (dStationCount < IncursionState)
                    {
                        if (MissionComponent.host.AiPay(DestroyerStation.BuildCosts))
                        {
                            if (MissionComponent.AiLivePing)
                            {
                                MissionComponent.Speak("Station manager", "Building Destroyer Station");
                            }
                            StartStationProduction(Station.StationType.Destroyer);
                        }
                    }
                    else if (sLStationCount - 1 < IncursionState)
                    {
                        
                        if (MissionComponent.host.AiPay(SmallFactoryStation.BuildCosts))
                        {
                            if (MissionComponent.AiLivePing)
                            {
                                MissionComponent.Speak("Station manager", "Building small factory Station");
                            }
                            StartStationProduction(Station.StationType.SmallLogistic);
                        }
                    }
                    else if (cStationCount - 2 < IncursionState)
                    {
                        if (MissionComponent.host.AiPay(CruiserStation.Buildconsts))
                        {
                            if (MissionComponent.AiLivePing)
                            {
                                MissionComponent.Speak("Station manager", "Building Cruiser Station");
                            }
                            StartStationProduction(Station.StationType.Cruiser);
                        }
                    }
                    else if (lMStationCount - 3 < IncursionState)
                    {
                        if (MissionComponent.host.AiPay(LargeLaunchPlatform.BuildCost))
                        {
                            if (MissionComponent.AiLivePing)
                            {
                                MissionComponent.Speak("Station manager", "Building Large missle Station");
                            }
                            StartStationProduction(Station.StationType.LargeLauncher);
                        }
                    }
                    else if (LLStationCount - 4 < IncursionState)
                    {
                        if (ecmStation == null)
                        {
                            if (MissionComponent.host.AiPay(800))
                            {
                                if (MissionComponent.AiLivePing)
                                {
                                    MissionComponent.Speak("Station manager", "Building ECM Station");
                                }
                                StartStationProduction(Station.StationType.Ecm);
                            }

                        }
                        else
                        {
                            if (MissionComponent.host.AiPay(LargeFactoryStation.BuildCosts))
                            {
                                if (MissionComponent.AiLivePing)
                                {
                                    MissionComponent.Speak("Station manager", "Building large factory Station");
                                }
                                StartStationProduction(Station.StationType.LargeLogistics);
                            }
                        }
                    }
                    else if (bStationCount - 5 < IncursionState)
                    {
                        if (MissionComponent.host.AiPay(BattleshipStation.BuildCosts))
                        {
                            if (MissionComponent.AiLivePing)
                            {
                                MissionComponent.Speak("Station manager", "Building battleship Station");
                            }
                            StartStationProduction(Station.StationType.Battleship);
                        }
                    }
                    else
                    {
                        if (ecmStation == null)
                        {
                            if (MissionComponent.host.AiPay(800))
                            {
                                StartStationProduction(Station.StationType.Ecm);
                            }
                        }

                        Random r = new Random(DateTime.Now.Minute);
                        int level = r.Next(1, 6);
                        switch (level)
                        {
                            case 1:
                                if (MissionComponent.host.AiPay(FrigateStation.BuildCosts))
                                {
                                    StartStationProduction(Station.StationType.Frigate);
                                }
                                break;
                            case 2:
                                if (MissionComponent.host.AiPay(DestroyerStation.BuildCosts))
                                {
                                    StartStationProduction(Station.StationType.Destroyer);
                                }
                                break;
                            case 3:
                                if (MissionComponent.host.AiPay(SmallFactoryStation.BuildCosts))
                                {
                                    StartStationProduction(Station.StationType.SmallLogistic);
                                }
                                break;
                            case 4:
                                if (MissionComponent.host.AiPay(CruiserStation.Buildconsts))
                                {
                                    StartStationProduction(Station.StationType.Cruiser);
                                }
                                break;
                            case 5:
                                if (MissionComponent.host.AiPay(LargeLaunchPlatform.BuildCost))
                                {
                                    StartStationProduction(Station.StationType.LargeLauncher);
                                }
                                break;
                            case 6:
                                if (MissionComponent.host.AiPay(LargeFactoryStation.BuildCosts))
                                {
                                    StartStationProduction(Station.StationType.LargeLogistics);
                                }
                                break;
                            case 7:
                                if (MissionComponent.host.AiPay(BattleshipStation.BuildCosts))
                                {
                                    StartStationProduction(Station.StationType.Battleship);
                                }
                                break;
                        }
                    }
                }
            }

            //start producing ships
            if (MissionComponent.writeDebug)
            {
                MissionComponent.Speak("Station manager", "buying ships");
            }
            foreach (var station in battleshipStation.Where(x => x.IsProducing == false))
            {
                var bsy = (BattleshipStation) station;
                if (MissionComponent.host.AiHasMoney(BattleShip.Price))
                {
                    bsy.StartProduction();
                }
                else
                {
                    break; //not enough money, go down a class
                }
            }

            foreach (var station in largeLauncherStation.Where(x => x.IsProducing == false))
            {
                var llp = (LargeLaunchPlatform) station;
                if (MissionComponent.host.AiHasMoney(750))
                {
                    llp.StartProduction();
                }
                else
                {
                    break; //stop the loop, ai does has less then 750 resources
                }
            }

            foreach (var station in cruiserStation.Where(x => x.IsProducing == false))
            {
                var cs = (CruiserStation) station;
                if (MissionComponent.host.AiHasMoney(Cruiser.Price))
                {
                    cs.StartProduction();
                }
                else
                {
                    break;
                }
            }
            foreach (var station in smallLauncherStation.Where(x => x.IsProducing == false))
            {
                var ss = (SmallLaunchPlatform) station;
                if (MissionComponent.host.AiHasMoney(40))
                {
                    ss.StartProduction();

                }
                else
                {
                    break;
                }
            }
            foreach (var station in destroyerStation.Where(x => x.IsProducing == false))
            {
                var ds = (DestroyerStation) station;
                if (MissionComponent.host.AiHasMoney(Destroyer.Price))
                {
                    ds.StartProduction();
                }
                else
                {
                    break;
                }
            }
            foreach (var station in frigateStations.Where(x => x.IsProducing == false))
            {
                var fs = (FrigateStation) station;
                if (MissionComponent.host.AiHasMoney(Frigate.Price))
                {
                    fs.StartProduction();
                }
                else
                {
                    break;
                }
            }
        }

        /// <summary>
        /// Get the incursion state
        /// </summary>
        /// <param name="destroyerStation"></param>
        /// <param name="smallLauncherStation"></param>
        /// <param name="smallLogisticStations"></param>
        /// <param name="cruiserStation"></param>
        /// <param name="largeLauncherStation"></param>
        /// <param name="ecmStation"></param>
        /// <param name="largeLogisticStations"></param>
        /// <param name="battleshipStations"></param>
        /// <param name="capitalStations"></param>
        /// <returns></returns>
        private int DeterminIncursionState(int destroyerStation,
            int smallLauncherStation,
            int cruiserStation, int largeLauncherStation, Station ecmStation,
            int largeLogisticStations, int battleshipStations,
            int capitalStations)
        {
            int incursionState = 0;
            if (destroyerStation > 0)
            {
                incursionState = 3;
            }
            else
            {
                incursionState = 2;
                return incursionState;
            }

            if (smallLauncherStation > 0)
            {
                incursionState = 4;
            }
            else
            {
                return incursionState;
            }

            if (cruiserStation > 0)
            {
                incursionState = 5;
            }
            else
            {
                return incursionState;
            }

            if (largeLauncherStation > 0)
            {
                incursionState = 6;
            }
            else
            {
                return incursionState;
            }

            if (ecmStation != null)
            {
                incursionState = 7;
            }
            else if (largeLogisticStations > 0)
            {
                incursionState = 7;
            }
            else
            {
                return incursionState;
            }

            if (battleshipStations > 0)
            {
                incursionState = 7;
            }
            else if (capitalStations > 0)
            {
                incursionState = 7;
            }
            if (MissionComponent.writeDebug)
            {
                MissionComponent.Speak("Incursion", "State is now " + incursionState);
            }
            return incursionState;
        }

        /// <summary>
        /// Start the production of a new station
        /// </summary>
        /// <param name="type">station type</param>
        private void StartStationProduction(Station.StationType type)
        {
            switch (type)
            {
                case Station.StationType.Frigate:

                    order = new ProductionOrder(5, type);

                    break;
                case Station.StationType.Destroyer:

                    order = new ProductionOrder(7, type);

                    break;
                case Station.StationType.SmallLauncher:

                    order = new ProductionOrder(7, type);

                    break;
                case Station.StationType.SmallLogistic:

                    order = new ProductionOrder(10, type);

                    break;
                case Station.StationType.Cruiser:

                    order = new ProductionOrder(15, type);

                    break;
                case Station.StationType.Battleship:
                    order = new ProductionOrder(30, type);
                    break;
                case Station.StationType.Ecm:
                    order = new ProductionOrder(17, type);
                    break;
                case Station.StationType.Capital:
                    order = new ProductionOrder(60, type);
                    break;
                case Station.StationType.LargeLauncher:
                    order = new ProductionOrder(25, type);
                    break;
                case Station.StationType.LargeLogistics:
                    order = new ProductionOrder(20, type);
                    break;
            }
            IsBuildingStation = true;
        }


    }
}
