﻿using System;
using System.Collections.Generic;
using System.Linq;
using Sandbox.ModAPI;
using Sandbox.ModAPI.Interfaces;
using VRage;
using VRageMath;

namespace Scripts.KSWH
{
    public static class ThreadWrapper
    {
        private static FastResourceLock mSpawnLock = new FastResourceLock();
        private static FastResourceLock mOrderLock = new FastResourceLock();
        private static List<SpawnOrder> mOrders = new List<SpawnOrder>(); //this is only for the spawn of a current cycle! 
        private static bool lost = false;
        private static PriorityQueue<int, SpawnOrder> SpawnQueue = new PriorityQueue<int, SpawnOrder>(); //this is for queueing up ready ships for a ai hook
        private static List<ActionOrder> mActionOrders = new List<ActionOrder>();
        public static FastResourceLock TargetLock = new FastResourceLock();

        public static void GetSpawnLock()
        {
            mSpawnLock.AcquireExclusive();
        }

        public static void ReleaseSpawnLock()
        {
            mSpawnLock.ReleaseExclusive();
        }

        /// <summary>
        /// deploy a new station
        /// NOTE: do not use this to spawn in AI ships! this call does NOT respect the Ai hook system implemented in the host data. Use the other constructor instead
        /// </summary>
        /// <param name="prefabName">name of the station to spawn</param>
        /// <param name="location">where should the station spawn</param>
        /// <param name="reciever">station hook that is producing the ship</param>
        public static void OrderSpawn(string prefabName, Vector3D location, Station reciever)
        {
            mOrderLock.AcquireExclusive();
            mOrders.Add(new SpawnOrder(prefabName, location, reciever));
            mOrderLock.ReleaseExclusive();
        }

        /// <summary>
        /// queue up a ship to spawn when a aihook is ready (heavy ships have priority)
        /// This call respects the Ai hook system
        /// </summary>
        /// <param name="prefabName">name of the ship to spawn</param>
        /// <param name="location">location of the ship to spawn</param>
        /// <param name="type">type of the ship that is spawning, used to determen priority and what AI hook to put on it</param>
        /// <param name="reference">Station with the build slot</param>
        public static void OrderSpawn(string prefabName, Vector3D location, Station.StationType type, Station reference)
        {
            mOrderLock.AcquireExclusive();
            //mOrders.Add(new SpawnOrder(prefabName, location, type));
            SpawnQueue.Add(new KeyValuePair<int, SpawnOrder>(DetermenPriority(type), new SpawnOrder(prefabName, location, type, reference)));
            mOrderLock.ReleaseExclusive();
        }

        /// <summary>
        /// Order a spawnn for something to pop in and give it a certain speed
        /// Note that this call will NOT supply the entity with a ai hook
        /// </summary>
        /// <param name="prefabName">name of rhe prefab to spawn</param>
        /// <param name="location">location of the prefab to spawn</param>
        /// <param name="speed">speed of the prefab</param>
        public static void OrderSpawn(string prefabName, Vector3D location, Vector3D speed)
        {
            mOrderLock.AcquireExclusive();
            mOrders.Add(new SpawnOrder(prefabName, location, speed));
            mOrderLock.ReleaseExclusive();
        }

        /// <summary>
        /// Numemr the priority (lower number means higher priority)
        /// </summary>
        /// <param name="s">the type</param>
        /// <returns>the priority (int)</returns>
        private static int DetermenPriority(Station.StationType s)
        {
            switch (s)
            {
                case Station.StationType.Frigate:
                    return 7;
                case Station.StationType.Destroyer:
                    return 6;
                case Station.StationType.SmallLauncher:
                    return 5;
                case Station.StationType.Cruiser:
                    return 4;
                case Station.StationType.Battleship:
                    return 2;
                case Station.StationType.Capital:
                    return 1;
                case Station.StationType.LargeLauncher:
                    return 3;
                default:
                    return 999;
            }
        }

        /// <summary>
        /// Struct to keep hold of the spawnorder data
        /// </summary>
        private struct SpawnOrder
        {
            public string PrefabName;
            public Vector3D Location;
            public Controlable Reciever;
            public Station.StationType OType;
            public bool IsShip;
            public Vector3D Speed;
            public bool AiUse;
            
            

            public SpawnOrder(string name, Vector3D location, Controlable reciever)
            {
                PrefabName = name;
                Location = location;
                Reciever = reciever;
                IsShip = false;
                OType = Station.StationType.Battleship;
                Speed = new Vector3D();
                AiUse = true;
            }

            public SpawnOrder(string name, Vector3D location, Station.StationType type, Station reciever)
            {
                PrefabName = name;
                Location = location;
                Reciever = reciever;
                OType = type;
                IsShip = true;
                Speed = new Vector3D();
                AiUse = true;
            }

            public SpawnOrder(string name, Vector3D location, Vector3D speed)
            {
                PrefabName = name;
                Location = location;
                Speed = speed;
                OType = Station.StationType.Battleship; //dummy placeholder
                IsShip = false;
                AiUse = false;
                Reciever = null;
            }
        }

        /// <summary>
        /// Wrapper object to hold a action that needs to be applied to a block that is not thread save (things that spawn in bullets)
        /// </summary>
        private struct ActionOrder
        {
            public List<IMySlimBlock> Blocks;
            public ITerminalAction Action;

            /// <summary>
            /// Execute the listed action
            /// </summary>
            public void ExecuteAction()
            {
                foreach (IMySlimBlock block in Blocks)
                {
                    try
                    {
                        Action.Apply(block.FatBlock);
                    }
                    catch (Exception)
                    {
                        //null pointer protection (aka the block got shot while the acction was applying
                    }
                    
                }
            }
        }

        /// <summary>
        /// Game over event (shell the player with warheads)
        /// </summary>
        public static void Armegeddon()
        {
            lost = true;
        }

        /// <summary>
        /// List a action event that needs to be applied to a series of blocks in thread savety
        /// </summary>
        /// <param name="blocks">list of blocks to apply the action to</param>
        /// <param name="action">the action that should be applied</param>
        public static void ListAction(List<IMySlimBlock> blocks, ITerminalAction action)
        {
            mActionOrders.Add(new ActionOrder{Action =  action, Blocks = blocks});
        }

        /// <summary>
        /// Causes a armegeddon in the game world (in other words it will start raining active warheads around the player)
        /// WARNING: this call is NOT threadsave
        /// </summary>
        private static void SpawnArmegeddon()
        {
            List<IMyPlayer> players = MissionComponent.GetPlayers();
            if (players == null || players.Count == 0)
            {
                //no player currently alive, don't spawn
                return;
            }
            Random r = new Random(DateTime.Now.Millisecond);
            Vector3D playerLoc = players[r.Next(players.Count - 1)].GetPosition();
            for (int i = 0; i < 5; i++)
            {
                double x = playerLoc.X + (r.Next(-40, 40));
                double y = playerLoc.Y + (r.Next(-40, 40));
                double z = playerLoc.Z + (r.Next(-40, 40));
                var spawns = Spawner.Spawn(StaticInformation.BomFile, new Vector3D(x, y, z));
                var tmp = new List<IMySlimBlock>();
                var bom = spawns[0] as IMyCubeGrid;
                bom.GetBlocks(tmp);
                var item = tmp.FirstOrDefault();
                var actions = new List<ITerminalAction>();
                MyAPIGateway.TerminalActionsHelper.GetActions(item.FatBlock.GetType(), actions, s => s.Id.Contains("Start"));
                foreach (ITerminalAction action in actions)
                {
                    action.Apply(item.FatBlock);
                }
            }
        }

        /// <summary>
        /// Handle that orders that need to be executed
        /// </summary>
        public static void HandleOrders()
        {
            mOrderLock.AcquireExclusive();
            try
            {
                //check if there are free hooks
                int freeHooks = MissionComponent.host.AiHookSlots - MissionComponent.host.ActiveHooks;
                if (freeHooks > 0)
                {
                    //free hooks, try to fill them
                    for (int i = 0; i < freeHooks; i++)
                    {
                        if (!SpawnQueue.IsEmpty)
                        {
                            var tmp = SpawnQueue.DequeueValue();
                            var stat = tmp.Reciever as Station;
                            if (stat != null && stat.Usefull)
                            {
                                //add ship to spawn orders if the station is still alive. Else forget about the order
                                mOrders.Add(tmp);
                            }
                        }
                        else
                        {
                            //no ships qued up, break and push on
                            break;
                        }
                    }
                }

                //spawn qued up orders (ships and stations)
                foreach (SpawnOrder order in mOrders)
                {
                    if (order.AiUse)
                    {
                        //spawn is used by the ai, its either a ship or a station
                        if (order.IsShip)
                        {
                            switch (order.OType)
                            {
                                case Station.StationType.Frigate:
                                    FrigateStation.HookAi(
                                        Spawner.Spawn(order.PrefabName, order.Location)[0] as IMyCubeGrid, MissionComponent.host.CreateshipId());
                                    break;
                                case Station.StationType.Destroyer:
                                    DestroyerStation.HookAi(
                                        Spawner.Spawn(order.PrefabName, order.Location)[0] as IMyCubeGrid, MissionComponent.host.CreateshipId());
                                    break;
                                case Station.StationType.SmallLauncher:
                                    SmallLaunchPlatform.HookAi(
                                        Spawner.Spawn(order.PrefabName, order.Location)[0] as IMyCubeGrid, MissionComponent.host.CreateshipId());
                                    break;
                                case Station.StationType.Cruiser:
                                    CruiserStation.HookAi(
                                        Spawner.Spawn(order.PrefabName, order.Location)[0] as IMyCubeGrid, MissionComponent.host.CreateshipId());
                                    break;
                                case Station.StationType.Battleship:
                                    BattleshipStation.HookAi(
                                        Spawner.Spawn(order.PrefabName, order.Location)[0] as IMyCubeGrid, MissionComponent.host.CreateshipId());
                                    break;
                                case Station.StationType.Capital:
                                    //todo: to be implemented
                                    break;
                                case Station.StationType.LargeLauncher:
                                    LargeLaunchPlatform.HookAi(
                                        Spawner.Spawn(order.PrefabName, order.Location)[0] as IMyCubeGrid, MissionComponent.host.CreateshipId());
                                    break;
                            }
                            try
                            {
                                ((Station) order.Reciever).FreeProductionSlot();
                            }
                            catch (Exception)
                            {
                                MissionComponent.Speak("Error", "Failed free production slot of station");
                            }

                            //clear the production slot so the ai can reuse it
                        }
                        else
                        {
                            order.Reciever.SetGrid(Spawner.Spawn(order.PrefabName, order.Location)[0] as IMyCubeGrid);
                        }
                    }
                    else
                    {
                        //ship is not ai used, its a ecm action
                        Spawner.SpawnWithSpeed(order.PrefabName, order.Location, order.Speed);
                    }
                }
                mOrders.Clear();
                if (lost)
                {
                    SpawnArmegeddon();
                }

                foreach (ActionOrder actionOrder in mActionOrders)
                {
                    actionOrder.ExecuteAction();
                }
                mActionOrders.Clear();
            }
            catch (Exception e)
            {
                MissionComponent.Speak("Error", "Failed to execute order: " + e);
            }
            mOrderLock.ReleaseExclusive();
        }

    }
}
